"""A wrapper that turns ordinary functions into Anthropic Tools."""
from __future__ import annotations
from dataclasses import dataclass
import inspect
from typing import Any, Callable, OrderedDict, TYPE_CHECKING, Type

from anthropic.types.beta.tools import ToolParam
from docstring_parser import Docstring, parse

from ..exceptions import BrokenSchemaError, CannotParseTypeError
from ..parsers import ArgSchemaParser, defargparsers

if TYPE_CHECKING:
    from ..json_type import JsonType


@dataclass
class WrapperConfig:
    """Configuration options for the ToolWrapper."""

    parsers: list[Type[ArgSchemaParser]] | None = None
    save_return: bool = True
    serialize: bool = True
    interpret_as_response: bool = False


class ToolWrapper:
    """A wrapper that turns ordinary functions into Anthropic Tools."""

    def __init__(
        self,
        func: Callable[..., Any],
        config: WrapperConfig | None = None,
        name: str | None = None,
        description: str | None = None,
    ) -> None:
        """Initialize the ToolWrapper.

        Args:
            func: The function to be wrapped.
            config: Configuration options for the wrapper.
            name: The name of the tool.
            description: The description of the tool.
        """
        self.func = func
        self.config = config or WrapperConfig()
        self._name = name
        self._description = description

    @property
    def parsers(self) -> list[Type[ArgSchemaParser]]:
        """Get the list of argument schema parsers.

        Returns:
            The list of argument schema parsers.
        """
        return self.config.parsers or defargparsers

    @property
    def save_return(self) -> bool:
        """Get the flag indicating whether to save the return value.

        Returns:
            The flag indicating whether to save the return value.
        """
        return self.config.save_return

    @property
    def serialize(self) -> bool:
        """Get the flag indicating whether to serialize the arguments.

        Returns:
            The flag indicating whether to serialize the arguments.
        """
        return self.config.serialize

    @property
    def interpret_as_response(self) -> bool:
        """Get the flag indicating whether to interpret the result as a response.

        Returns:
            The flag indicating whether to interpret the result as a response.
        """
        return self.config.interpret_as_response

    @property
    def argument_parsers(self) -> OrderedDict[str, ArgSchemaParser]:
        """Get the ordered dictionary of argument parsers.

        Returns:
            The ordered dictionary of argument parsers.
        """
        return OrderedDict(
            (
                (name, self.parse_argument(argument))
                for name, argument in inspect.signature(self.func).parameters.items()
            )
        )

    @property
    def required_arguments(self) -> list[str]:
        """Get the list of required arguments.

        Returns:
            The list of required arguments.
        """
        return [
            name
            for name, argument in inspect.signature(self.func).parameters.items()
            if argument.default is argument.empty
        ]

    @property
    def arguments_schema(self) -> JsonType:
        """Get the schema for the arguments.

        Returns:
            The schema for the arguments.
        """
        return {
            name: {
                **parser.argument_schema,
                **(
                    {"description": self.arg_docs.get(name)}
                    if name in self.arg_docs
                    else {}
                ),
            }
            for name, parser in self.argument_parsers.items()
        }

    @property
    def parsed_docs(self) -> Docstring:
        """Get the parsed docstring.

        Returns:
            The parsed docstring.
        """
        return parse(self.func.__doc__ or "")

    @property
    def arg_docs(self) -> dict[str, str]:
        """Get the dictionary of argument descriptions.

        Returns:
            The dictionary of argument descriptions.
        """
        return {
            param.arg_name: param.description
            for param in self.parsed_docs.params
            if param.description
        }

    @property
    def name(self) -> str:
        """Get the name of the tool.

        Returns:
            The name of the tool.
        """
        return self._name or self.func.__name__

    @property
    def schema(self) -> ToolParam:
        """Get the schema for the tool.

        Returns:
            The schema for the tool.
        """
        schema: ToolParam = {
            "name": self.name,
            "input_schema": {
                "type": "object",
                "properties": self.arguments_schema,
                "required": self.required_arguments,  # type: ignore
            },
        }
        description = self.parsed_docs.short_description or self._description
        if description:
            schema["description"] = description
        return schema

    def parse_argument(self, argument: inspect.Parameter) -> ArgSchemaParser:
        """Parse the argument using the appropriate argument schema parser.

        Args:
            argument: The argument to be parsed.

        Returns:
            The argument schema parser.

        Raises:
            CannotParseTypeError: If the argument type is not supported.
        """
        for parser in self.parsers:
            if parser.can_parse(argument.annotation):
                return parser(argument.annotation, self.parsers)
        raise CannotParseTypeError(argument.annotation)

    def parse_arguments(self, arguments: dict[str, JsonType]) -> OrderedDict[str, Any]:
        """Parse the arguments using the argument parsers.

        Args:
            arguments: The arguments to be parsed.

        Returns:
            The ordered dictionary of parsed arguments.

        Raises:
            BrokenSchemaError: If the arguments do not match the schema.
        """
        argument_parsers = self.argument_parsers
        if not all(name in arguments for name in self.required_arguments):
            raise BrokenSchemaError(arguments, self.arguments_schema)
        try:
            return OrderedDict(
                (
                    (name, argument_parsers[name].parse_value(value))
                    for name, value in arguments.items()
                )
            )
        except KeyError as err:
            raise BrokenSchemaError(arguments, self.arguments_schema) from err

    def __call__(self, arguments: dict[str, JsonType]) -> Any:
        """Call the wrapped function with the parsed arguments.

        Args:
            arguments: The arguments to be passed to the function.

        Returns:
            The result of calling the function.
        """
        return self.func(**self.parse_arguments(arguments))



"""Anthropic Claude tool use based on function docstrings."""
from .conversation import Conversation
from .exceptions import (
    AnthropicToolsError,
    BrokenSchemaError,
    CannotParseTypeError,
    NonSerializableOutputError,
    ToolNotFoundError,
)
from .parsers import ArgSchemaParser, defargparsers
from .tools import (
    AnthropicTool,
    BasicToolSet,
    ImageToolOutput,
    MutableToolSet,
    RawToolResult,
    TextToolOutput,
    ToolOutput,
    ToolResult,
    ToolSet,
    ToolWrapper,
    UnionSkillSet,
    WrapperConfig,
)

__all__ = [
    "AnthropicTool",
    "AnthropicToolsError",
    "ArgSchemaParser",
    "BasicToolSet",
    "BrokenSchemaError",
    "CannotParseTypeError",
    "Conversation",
    "ImageToolOutput",
    "MutableToolSet",
    "NonSerializableOutputError",
    "RawToolResult",
    "TextToolOutput",
    "ToolNotFoundError",
    "ToolOutput",
    "ToolResult",
    "ToolSet",
    "ToolWrapper",
    "UnionSkillSet",
    "WrapperConfig",
    "defargparsers",
]



"""Parser for null types"""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Type

from ..exceptions import BrokenSchemaError
from .parser import ArgSchemaParser

if TYPE_CHECKING:
    from ..json_type import JsonType
    from typing_extensions import TypeGuard


class NoneParser(ArgSchemaParser[None]):
    """Parser for null types"""

    @property
    def argument_schema(self) -> dict[str, JsonType]:
        return {"type": "null"}

    @classmethod
    def can_parse(cls, argtype: Any) -> TypeGuard[Type[None]]:
        return argtype in [None, type(None)]

    def parse_value(self, value: JsonType) -> None:
        if value is not None:
            raise BrokenSchemaError(value, self.argument_schema)



"""Parser for dict types"""
from __future__ import annotations
from typing import Any, Dict, TYPE_CHECKING, Type, TypeVar, get_args, get_origin

from ..exceptions import BrokenSchemaError
from .parser import ArgSchemaParser

if TYPE_CHECKING:
    from ..json_type import JsonType
    from typing_extensions import TypeGuard

T = TypeVar("T")


class DictParser(ArgSchemaParser[Dict[str, T]]):
    """Parser for dict types"""

    @property
    def argument_schema(self) -> Dict[str, JsonType]:
        return {
            "type": "object",
            "additionalProperties": self.parse_rec(
                get_args(self.argtype)[1]
            ).argument_schema,
        }

    @classmethod
    def can_parse(cls, argtype: Any) -> TypeGuard[Type[Dict[str, T]]]:
        return (
            get_origin(argtype)
            in [
                dict,
                Dict,
            ]
            and get_args(argtype)[0] is str
        )

    def parse_value(self, value: JsonType) -> Dict[str, T]:
        if not isinstance(value, dict):
            raise BrokenSchemaError(value, self.argument_schema)
        return {
            k: self.parse_rec(get_args(self.argtype)[1]).parse_value(v)
            for k, v in value.items()
        }



"""A protocol for a parser for a specific argument type."""
from __future__ import annotations
from typing import Any, Protocol, TYPE_CHECKING, Type, TypeVar

from ..exceptions import CannotParseTypeError

if TYPE_CHECKING:
    from ..json_type import JsonType
    from typing_extensions import TypeGuard

T = TypeVar("T")
S = TypeVar("S")


class ArgSchemaParser(Protocol[T]):
    """A protocol for a parser for a specific argument type

    Both converts the argument definition to a JSON schema and parses the argument value
    from JSON import
    """

    argtype: Type[T]
    rec_parsers: list[Type[ArgSchemaParser]]

    def __init__(
        self, argtype: Type[T], rec_parsers: list[Type[ArgSchemaParser]]
    ) -> None:
        self.argtype = argtype
        self.rec_parsers = rec_parsers

    @property
    def argument_schema(self) -> dict[str, JsonType]:
        """Parse an argument of a specific type"""
        ...  # pylint: disable=unnecessary-ellipsis

    @classmethod
    def can_parse(cls, argtype: Any) -> TypeGuard[Type[T]]:
        """Whether this parser can parse a specific arg type

        Args:
            argtype (Any): The type to check
        """
        ...  # pylint: disable=unnecessary-ellipsis

    def parse_rec(self, argtype: Type[S]) -> ArgSchemaParser[S]:
        """Parse a type recursively

        Args:
            argtype (Type[S]): The type to parse

        Returns:
            ArgSchemaParser[S]: The parser for the type

        Raises:
            CannotParseTypeError: If the type cannot be parsed
        """
        for parser in self.rec_parsers:
            if parser.can_parse(argtype):
                return parser(argtype, self.rec_parsers)
        raise CannotParseTypeError(argtype)

    def parse_value(self, value: JsonType) -> T:
        """Parse a value of a specific type

        Args:
            value (JsonType): The value to parse

        Raises:
            BrokenSchemaError: If the value does not match the schema
        """
        ...  # pylint: disable=unnecessary-ellipsis



"""A set of tools responsible for managing the skill sets."""
from .basic_set import BasicToolSet
from .sets import AnthropicTool, MutableToolSet, ToolSet
from .tools import (
    ImageToolOutput,
    RawToolResult,
    TextToolOutput,
    ToolOutput,
    ToolResult,
)
from .union import UnionSkillSet
from .wrapper import ToolWrapper, WrapperConfig

__all__ = [
    "AnthropicTool",
    "BasicToolSet",
    "ImageToolOutput",
    "MutableToolSet",
    "RawToolResult",
    "TextToolOutput",
    "ToolOutput",
    "ToolResult",
    "ToolSet",
    "ToolWrapper",
    "UnionSkillSet",
    "WrapperConfig",
]



"""Parser for bool types"""
from .atomic_type_parser import AtomicParser


class BoolParser(AtomicParser[bool]):
    """Parser for bool types"""

    _type = bool
    schema_type_name: str = "boolean"



"""Parser for int types"""
from __future__ import annotations
from typing import TYPE_CHECKING

from ..exceptions import BrokenSchemaError
from .atomic_type_parser import AtomicParser

if TYPE_CHECKING:
    from ..json_type import JsonType


class IntParser(AtomicParser[int]):
    """Parser for int types"""

    _type = int
    schema_type_name: str = "integer"

    def parse_value(self, value: JsonType) -> int:
        if isinstance(value, bool):
            # This has to happen for historical reasons
            # bool is a subclass of int, so isinstance(value, int) is True
            raise BrokenSchemaError(value, self.argument_schema)
        return super().parse_value(value)



"""A definition of a tool set that can be used by Anthropic Claude."""
from __future__ import annotations
from dataclasses import dataclass
import json
from typing import (
    Any,
    Iterable,
    Literal,
    Protocol,
    TYPE_CHECKING,
    TypeGuard,
    runtime_checkable,
)

from anthropic.types import ImageBlockParam, TextBlockParam
from anthropic.types.beta.tools import (
    ToolResultBlockParam,
    ToolUseBlockParam,
    ToolsBetaContentBlock,
)
from anthropic.types.image_block_param import Source

from ..exceptions import NonSerializableOutputError

if TYPE_CHECKING:
    from anthropic.types.beta.tools import ToolParam
    from ..json_type import JsonType


@runtime_checkable
class AnthropicTool(Protocol):
    """Protocol for Anthropic Tools.

    Anthropic Tools are callable objects that can be used by Anthropic Claude,
    what sets them apart from a regular function is that they have a schema.
    """

    def __call__(self, arguments: dict[str, JsonType]) -> Any:
        """Call the Anthropic Tool with the given arguments.

        Args:
            arguments: A dictionary of arguments to be passed to the Anthropic Tool.

        Returns:
            The result of the Anthropic Tool.

        Raises:
            Any exception raised by the Anthropic Tool.
        """

    @property
    def schema(self) -> ToolParam:
        """Get the schema of the Anthropic Tool.

        Returns:
            The schema of the Anthropic Tool.
        """
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def name(self) -> str:
        """Get the name of the Tool.

        Returns:
            The name of the Tool.
        """
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def save_return(self) -> bool:
        """Check if the return value of the Tool should be added to the convo."""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def serialize(self) -> bool:
        """Check if the return value of the Tool should be serialized."""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def interpret_as_response(self) -> bool:
        """Check if the return value of the Anthropic Tool should be interpreted as
        the assistant's response.
        """
        ...  # pylint: disable=unnecessary-ellipsis


@dataclass
class TextToolOutput:
    """The text output of a tool."""

    content: str
    is_error: bool = False
    type: Literal["text"] = "text"


@dataclass
class ImageToolOutput:
    """The image output of a tool."""

    source: Source
    type: Literal["image"] = "image"


ToolOutput = TextToolOutput | ImageToolOutput


@dataclass
class RawToolResult:
    """The raw result of an Anthropic Tool.

    The raw result can be serialized and converted to content blocks.
    """

    result: Any
    serialize: bool = True

    @property
    def serialized(self) -> str:
        """Serialize the raw result.

        Returns:
            The serialized raw result.

        Raises:
            NonSerializableOutputError: If the raw result cannot be serialized.
        """
        if self.serialize:
            try:
                return json.dumps(self.result, indent=2)
            except TypeError as error:
                raise NonSerializableOutputError(self.result) from error
        return str(self.result)

    @property
    def as_content(
        self,
    ) -> (
        str
        | Iterable[
            ToolsBetaContentBlock
            | TextBlockParam
            | ImageBlockParam
            | ToolUseBlockParam
            | ToolResultBlockParam
        ]
    ):
        """Convert the raw result to content blocks.

        Returns:
            The content blocks representing the raw result.
        """
        if self.serialize:
            return self.serialized
        return self.result


@dataclass
class ToolResult:
    """Represents the result of using an Anthropic Tool.

    The result can include the content blocks and the interpreted return value.
    """

    use_id: str
    name: str
    raw_result: RawToolResult
    interpret_return_as_response: bool = False
    response_role: Literal["user", "assistant"] = "user"

    @property
    def content(self) -> str:
        """Get the content of the Tool Result.

        Returns:
            The content of the Tool Result, or None if there is no raw result.
        """
        return self.raw_result.serialized

    @property
    def blocks(
        self,
    ) -> (
        str
        | Iterable[
            ToolsBetaContentBlock
            | TextBlockParam
            | ImageBlockParam
            | ToolUseBlockParam
            | ToolResultBlockParam
        ]
        | None
    ):
        """Get the content blocks of the Tool Result. Only meaningful for tools
        that have interpret_return_as_response set to True. An example use case
        is a tool that returns image content blocks.

        Returns:
            The content blocks of the Tool Result,
                or None if there is no raw result.
        """
        result = self.raw_result.as_content
        if self._is_tool_output(result):
            blocks = [
                ToolResultBlockParam(
                    type="tool_result",
                    tool_use_id=self.use_id,
                    content=[{"type": "text", "text": block.content}],
                    is_error=block.is_error,
                )
                if block.type == "text"
                else ImageBlockParam(type="image", source=block.source)
                for block in result
            ]
            if not any(block["type"] == "tool_result" for block in blocks):
                blocks.insert(
                    0,
                    ToolResultBlockParam(
                        type="tool_result",
                        tool_use_id=self.use_id,
                    ),
                )
            if len([block for block in blocks if block["type"] == "tool_result"]) > 1:
                # Replace the first result block with all concatenated results:
                first_text_block = next(
                    block for block in blocks if block["type"] == "tool_result"
                )
                first_text_block["content"] = sum(
                    [
                        list(block.get("content", []))
                        for block in blocks
                        if block["type"] == "tool_result"
                    ],
                    [],
                )
                first_text_block["is_error"] = any(
                    block["is_error"]
                    for block in blocks
                    if block["type"] == "tool_result"
                )
                # Delete all but the first one now:
                new_blocks = []
                first_done = False
                for block in blocks:
                    if block["type"] == "tool_result":
                        if first_done:
                            continue
                        else:
                            first_done = True
                    new_blocks.append(first_text_block)
                return new_blocks
            return blocks
        return result

    @property
    def result(self) -> Any | None:
        """Get the result of the underlying function call.

        Returns:
            The raw result of the function call
        """
        if self.raw_result:
            return self.raw_result.result
        return None

    def _is_tool_output(self, result: Any) -> TypeGuard[list[ToolOutput]]:
        """Check if the result is a list of ToolOutput objects.

        Args:
            result: The result to check.

        Returns:
            True if the result is a list of ToolOutput objects,
                False otherwise.
        """
        return isinstance(result, list) and all(
            isinstance(item, (TextToolOutput, ImageToolOutput)) for item in result
        )



"""Protocols for the sets of tools used by Anthropic Claude."""
from __future__ import annotations
from functools import partial
from typing import Any, Callable, Protocol, TYPE_CHECKING, overload

from anthropic.types.beta.tools import ToolParam, ToolUseBlock, ToolUseBlockParam

from .tools import AnthropicTool, ToolResult
from .wrapper import ToolWrapper, WrapperConfig

if TYPE_CHECKING:
    from ..json_type import JsonType


class ToolSet(Protocol):
    """A protocol defining a set of tools that can be used by Anthropic Claude."""

    @property
    def tools_schema(self) -> list[ToolParam]:
        """Get the schema of the tools in the set."""
        ...  # pylint: disable=unnecessary-ellipsis

    def run_tool(self, input_data: ToolUseBlock | ToolUseBlockParam) -> ToolResult:
        """Run a tool from the set with the given input data.

        Args:
            input_data: The input data for the tool.
        """
        ...  # pylint: disable=unnecessary-ellipsis

    def __call__(self, input_data: ToolUseBlock | ToolUseBlockParam) -> JsonType:
        """Call the tool set with the given input data.

        Args:
            input_data: The input data for the tool.

        Returns:
            The result of running the tool.
        """
        return self.run_tool(input_data).result


class MutableToolSet(ToolSet, Protocol):
    """A protocol defining a mutable set of tools that can be used by Anthropic Claude."""

    def _add_tool(self, tool: AnthropicTool) -> None:
        """Add a tool to the set.

        Args:
            tool: The tool to add.
        """
        ...  # pylint: disable=unnecessary-ellipsis

    @overload
    def add_tool(self, tool: AnthropicTool) -> AnthropicTool:
        ...  # pylint: disable=unnecessary-ellipsis

    @overload
    def add_tool(
        self,
        tool: Callable[..., Any],
        *,
        name: str | None = None,
        description: str | None = None,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[..., Any]:
        ...  # pylint: disable=unnecessary-ellipsis

    @overload
    def add_tool(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        ...  # pylint: disable=unnecessary-ellipsis

    def add_tool(
        self,
        tool: AnthropicTool | Callable[..., Any] | None = None,
        *,
        name: str | None = None,
        description: str | None = None,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]] | Callable[..., Any]:
        """Add a tool to the set.

        Args:
            tool: The tool to add.
            name: The name of the tool.
            description: The description of the tool.
            save_return: Whether to save the return value of the tool.
            serialize: Whether to serialize the return value of the tool.
            interpret_as_response: Whether to interpret the return value as a response.

        Returns:
            The added tool or a decorator to add a callable tool.
        """
        if isinstance(tool, AnthropicTool):
            self._add_tool(tool)
            return tool
        if callable(tool):
            self._add_tool(
                ToolWrapper(
                    tool,
                    WrapperConfig(None, save_return, serialize, interpret_as_response),
                    name=name,
                    description=description,
                )
            )
            return tool
        return partial(
            self.add_tool,
            name=name,
            description=description,
            save_return=save_return,
            serialize=serialize,
            interpret_as_response=interpret_as_response,
        )

    def remove_tool(self, tool: str | AnthropicTool | Callable[..., Any]) -> None:
        """Remove a tool from the set.

        Args:
            tool: The tool to remove. It can be the name of the tool,
                an AnthropicTool, or a callable.
        """
        if isinstance(tool, str):
            self._remove_tool(tool)
            return
        if isinstance(tool, AnthropicTool):
            self._remove_tool(tool.name)
            return
        self._remove_tool(tool.__name__)

    def _remove_tool(self, name: str) -> None:
        ...



"""Parser for atomic json types"""
from __future__ import annotations
from typing import Any, Protocol, TYPE_CHECKING, Type, TypeVar

from ..exceptions import BrokenSchemaError
from .parser import ArgSchemaParser

if TYPE_CHECKING:
    from ..json_type import JsonType
    from typing_extensions import TypeGuard

T = TypeVar("T")


class AtomicParser(ArgSchemaParser[T], Protocol[T]):
    """Parser for atomic json values"""

    _type: Type[T]

    @property
    def schema_type_name(self) -> str:
        """Name of the type in the json schema"""
        ...  # pylint: disable=unnecessary-ellipsis

    @property
    def argument_schema(self) -> dict[str, JsonType]:
        return {
            "type": self.schema_type_name,
        }

    @classmethod
    def can_parse(cls, argtype: Any) -> TypeGuard[Type[T]]:
        return argtype is cls._type

    def parse_value(self, value: JsonType) -> T:
        if not isinstance(value, self._type):
            raise BrokenSchemaError(value, self.argument_schema)
        return value



"""Parser for float types"""
from __future__ import annotations
from typing import TYPE_CHECKING

from ..exceptions import BrokenSchemaError
from .atomic_type_parser import AtomicParser

if TYPE_CHECKING:
    from ..json_type import JsonType


class FloatParser(AtomicParser[float]):
    """Parser for float types"""

    _type = float
    schema_type_name: str = "number"

    def parse_value(self, value: JsonType) -> float:
        if not isinstance(value, (float, int)):
            raise BrokenSchemaError(value, self.argument_schema)
        return float(value)



"""Parser for string types"""
from .atomic_type_parser import AtomicParser


class StringParser(AtomicParser[str]):
    """Parser for string types"""

    _type = str
    schema_type_name: str = "string"



"""A skill set that combines multiple tool sets."""
from __future__ import annotations
import contextlib
from typing import TYPE_CHECKING

from anthropic.types.beta.tools import ToolParam, ToolUseBlock, ToolUseBlockParam

from ..exceptions import ToolNotFoundError
from .basic_set import BasicToolSet

if TYPE_CHECKING:
    from .tools import ToolResult
    from .sets import ToolSet


class UnionSkillSet(BasicToolSet):
    """A skill set that combines multiple tool sets."""

    def __init__(self, *sets: ToolSet) -> None:
        """Initialize the UnionSkillSet.

        Args:
            sets: The tool sets to be combined.
        """
        self.sets = list(sets)
        super().__init__()

    @property
    def tools_schema(self) -> list[ToolParam]:
        """Get the schema of the tools in the skill set.

        Returns:
            The list of tool parameters.
        """
        return super().tools_schema + sum(
            (tool_set.tools_schema for tool_set in self.sets), []
        )

    def run_tool(self, input_data: ToolUseBlock | ToolUseBlockParam) -> ToolResult:
        """Run a tool using the skill set.

        Args:
            input_data: The input data for the tool.

        Returns:
            The result of running the tool.
        """
        for tool_set in self.sets:
            with contextlib.suppress(ToolNotFoundError):
                return tool_set.run_tool(input_data)
        return super().run_tool(input_data)

    def add_skill(self, skill: ToolSet) -> None:
        """Add a tool set to the skill set.

        Args:
            skill: The tool set to be added.
        """
        self.sets.append(skill)



"""The exceptions associated with Anthropic Tools."""

from __future__ import annotations
from typing import Any


class AnthropicToolsError(Exception):
    """Base class for exceptions raised by Anthropic Tools."""


class ToolNotFoundError(AnthropicToolsError):
    """The requested tool is not found.

    Attributes:
        tool_name (str): The name of the tool that was not found.
    """

    def __init__(self, tool_name: str) -> None:
        """Initialize the ToolNotFoundError.

        Args:
            tool_name (str): The name of the tool that was not found.
        """
        super().__init__(f"Tool {tool_name} not found.")
        self.tool_name = tool_name


class CannotParseTypeError(AnthropicToolsError):
    """The type of an argument cannot be parsed.

    Attributes:
        argtype (Any): The type that cannot be parsed.
    """

    def __init__(self, argtype: Any) -> None:
        """Initialize the CannotParseTypeError.

        Args:
            argtype (Any): The type that cannot be parsed.
        """
        super().__init__(f"Cannot parse type {argtype}")
        self.argtype = argtype


class NonSerializableOutputError(AnthropicToolsError):
    """The tool output is not JSON-serializable.

    Attributes:
        result (Any): The result that is not JSON-serializable.
    """

    def __init__(self, result: Any) -> None:
        """Initialize the NonSerializableOutputError.

        Args:
            result (Any): The result that is not JSON-serializable.
        """
        super().__init__(
            f"The result {result} is not JSON-serializable. "
            "Set serialize=False to use str() instead."
        )
        self.result = result


class BrokenSchemaError(AnthropicToolsError):
    """The response from Anthropic does not match the schema.

    Attributes:
        response (Any): The response that does not match the schema.
        schema (Any): The schema that the response should match.
    """

    def __init__(self, response: Any, schema: Any) -> None:
        """Initialize the BrokenSchemaError.

        Args:
            response (Any): The response that does not match the schema.
            schema (Any): The schema that the response should match.
        """
        super().__init__(
            "Anthropic returned a response that did not match the schema: "
            f"{response!r} does not match {schema}"
        )
        self.response = response
        self.schema = schema



"""A conversation with Anthropic Claude."""
from __future__ import annotations
import time
from typing import Any, Callable, TYPE_CHECKING, overload

from anthropic import Anthropic, RateLimitError
from anthropic.types import TextBlockParam
from anthropic.types.beta.tools import (
    ToolParam,
    ToolResultBlockParam,
    ToolsBetaMessage,
    ToolsBetaMessageParam,
)

from .tools.union import UnionSkillSet

if TYPE_CHECKING:
    from .tools import AnthropicTool
    from .tools import ToolResult, ToolSet


class Conversation:
    """A conversation with Anthropic Claude with tools."""

    def __init__(
        self,
        client: Anthropic,
        skills: list[ToolSet] | None = None,
        model: str = "claude-3-haiku-20240307",
        engine: str | None = None,
    ) -> None:
        """Initialize the Conversation class.

        Args:
            client (Anthropic): The Anthropic client.
            skills (list[ToolSet] | None, optional): The list of tool sets to use.
                Defaults to None.
            model (str, optional): The model to use for generating messages.
                Defaults to "claude-3-haiku-20240307".
            engine (str | None, optional): The engine to use. Defaults to None.
        """
        self.client = client
        self.messages: list[ToolsBetaMessageParam] = []
        self.skills = UnionSkillSet(*(skills or []))
        self.model = model
        self.engine = engine

    @property
    def tools_schema(self) -> list[ToolParam]:
        """Get the tools schema.

        Returns:
            list[ToolParam]: The tools schema.
        """
        return self.skills.tools_schema

    def add_message(self, message: ToolsBetaMessageParam | str) -> None:
        """Add a message to the conversation.

        Args:
            message (ToolsBetaMessageParam | str): The message to add.
        """
        if isinstance(message, str):
            self._add_message({"role": "user", "content": message})
        else:
            self._add_message(message)

    def add_messages(self, messages: list[ToolsBetaMessageParam]) -> None:
        """Add multiple messages to the conversation.

        Args:
            messages (list[ToolsBetaMessageParam]): The messages to add.
        """
        for message in messages:
            self.add_message(message)

    def pop_message(self, index: int = -1) -> ToolsBetaMessageParam:
        """Pop a message

        Args:
            index (int): The index. Defaults to -1.

        Returns:
            ToolsBetaMessageParam: The popped message.
        """
        return self.messages.pop(index)

    def clear_messages(self) -> None:
        """Clear all messages from the conversation."""
        self.messages = []

    def _add_message(self, message: ToolsBetaMessageParam) -> None:
        """Add a message to the conversation.

        Args:
            message (ToolsBetaMessageParam): The message to add.
        """
        self.messages.append(message)

    def _generate_message(
        self, retries: int | None = 1, max_tokens: int = 1024
    ) -> ToolsBetaMessage:
        """Generate a message using Anthropic Claude.

        Args:
            retries (int | None, optional): The number of retries. Defaults to 1.
            max_tokens (int, optional): The maximum number of tokens. Defaults to 1024.

        Returns:
            ToolsBetaMessage: The generated message.
        """
        if retries is None:
            retries = -1
        while True:
            try:
                response = self._generate_raw_message(max_tokens=max_tokens)
            except RateLimitError:
                if retries == 0:
                    raise
                retries -= 1
                time.sleep(1)
            else:
                return response

    def _generate_raw_message(self, max_tokens: int = 1024) -> ToolsBetaMessage:
        """Generate a raw message using Anthropic Claude.

        Args:
            max_tokens (int, optional): The maximum number of tokens. Defaults to 1024.

        Returns:
            ToolsBetaMessage: The generated raw message.
        """
        self._merge_same_role_messages()
        return self.client.beta.tools.messages.create(
            model=self.model,
            max_tokens=max_tokens,
            tools=self.tools_schema,
            messages=self.messages,
        )

    def _merge_same_role_messages(self) -> None:
        """Merge messages with the same role if they are adjacent."""
        if not self.messages:
            return
        merged_messages = [self.messages[0]]
        for message in self.messages[1:]:
            if message["role"] == merged_messages[-1]["role"]:
                if isinstance(merged_messages[-1]["content"], str):
                    merged_messages[-1]["content"] = [
                        TextBlockParam(type="text", text=merged_messages[-1]["content"])
                    ]
                if isinstance(message["content"], str):
                    message["content"] = [
                        TextBlockParam(type="text", text=message["content"])
                    ]
                merged_messages[-1]["content"].extend(message["content"])
            else:
                merged_messages.append(message)
        self.messages = merged_messages

    def add_tool_result(self, tool_result: ToolResult) -> bool:
        """Add a tool result to the conversation.

        Args:
            tool_result (ToolResult): The tool result to add.

        Returns:
            bool: True if the tool result was added, False otherwise.
        """
        if tool_result.interpret_return_as_response:
            self.add_message(
                {
                    "role": tool_result.response_role,
                    "content": tool_result.blocks or tool_result.content,
                }
            )
        else:
            self.add_message(
                ToolsBetaMessageParam(
                    role=tool_result.response_role,
                    content=[
                        ToolResultBlockParam(
                            tool_use_id=tool_result.use_id,
                            type="tool_result",
                            content=[
                                TextBlockParam(type="text", text=tool_result.content)
                            ],
                            is_error=False,
                        )
                    ],
                )
            )
        return True

    def run_tool_if_needed(self) -> bool:
        """Run a tool if needed.

        Returns:
            bool: True if a tool was run, False otherwise.
        """
        if not self.messages:
            return False
        last_message = self.messages[-1]
        ran = False
        for block in last_message["content"]:
            if isinstance(block, str):
                continue
            if (block["type"] if isinstance(block, dict) else block.type) == "tool_use":
                # We know that the block is a ToolUseBlockParam
                self.add_tool_result(self.skills.run_tool(input_data=block))  # type: ignore
                ran = True
        return ran

    def generate_message(
        self, retries: int | None = 1, max_tokens=1024
    ) -> ToolsBetaMessageParam:
        """Generate a message using Anthropic Claude.

        Args:
            retries (int | None, optional): The number of retries. Defaults to 1.
            max_tokens (int, optional): The maximum number of tokens. Defaults to 1024.

        Returns:
            ToolsBetaMessageParam: The generated message.
        """
        if self.run_tool_if_needed():
            return self.messages[-1]
        message: ToolsBetaMessage = self._generate_message(retries, max_tokens)
        self.add_message({"role": message.role, "content": message.content})
        return self.messages[-1]

    def _contains_call(self, message: ToolsBetaMessageParam) -> bool:
        """Check if a message contains a tool call.

        Args:
            message (ToolsBetaMessageParam): The message to check.

        Returns:
            bool: True if the message contains a tool call, False otherwise.
        """
        for block in message["content"]:
            if isinstance(block, str):
                continue
            if (block["type"] if isinstance(block, dict) else block.type) == "tool_use":
                return True
        return False

    def run_until_response(
        self, retries: int | None = 1, max_tokens=1024
    ) -> list[ToolsBetaMessageParam]:
        """Run the conversation until a response is received.

        Args:
            retries (int | None, optional): The number of retries. Defaults to 1.
            max_tokens (int, optional): The maximum number of tokens per message.
                Defaults to 1024.

        Returns:
            list[ToolsBetaMessageParam]: The response messages,
                both from Claude and tools.
        """
        generated = []
        while True:
            message = self.generate_message(retries=retries, max_tokens=max_tokens)
            generated.append(message)
            if not self._contains_call(message):
                return generated

    @overload
    def add_tool(self, tool: AnthropicTool) -> AnthropicTool:
        ...

    @overload
    def add_tool(
        self,
        tool: Callable[..., Any],
        *,
        name: str | None = None,
        description: str | None = None,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[..., Any]:
        ...

    @overload
    def add_tool(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        ...

    def add_tool(
        self,
        tool: AnthropicTool | Callable[..., Any] | None = None,
        *,
        name: str | None = None,
        description: str | None = None,
        save_return: bool = True,
        serialize: bool = True,
        interpret_as_response: bool = False,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]] | Callable[..., Any]:
        """Add a tool to the conversation.

        Args:
            tool (AnthropicTool | Callable[..., Any] | None, optional): The tool to add.
                Defaults to None.
            name (str | None, optional): The name of the tool. Defaults to None.
            description (str | None, optional): The description of the tool.
                Defaults to None.
            save_return (bool, optional): Whether to save the return value of the tool.
                Defaults to True.
            serialize (bool, optional): Whether to serialize the tool. Defaults to True.
            interpret_as_response (bool, optional): Whether to interpret the tool
                as a response. Defaults to False.

        Returns:
            Callable[[Callable[..., Any]], Callable[..., Any]] | Callable[..., Any]:
                The added tool, or a decorator that adds the tool.
        """
        if tool is None:
            return self.skills.add_tool(
                name=name,
                description=description,
                save_return=save_return,
                serialize=serialize,
                interpret_as_response=interpret_as_response,
            )
        return self.skills.add_tool(
            tool,
            name=name,
            description=description,
            save_return=save_return,
            serialize=serialize,
            interpret_as_response=interpret_as_response,
        )

    def remove_tool(self, tool: str | AnthropicTool | Callable[..., Any]) -> None:
        """Remove a tool from the conversation.

        Args:
            tool (str | AnthropicTool | Callable[..., Any]): The tool to remove.
        """
        self.skills.remove_tool(tool)

    def ask(
        self, question: str, retries: int | None = 1, max_tokens: int = 1024
    ) -> list[ToolsBetaMessageParam]:
        """Ask a question to Anthropic Claude.

        Args:
            question (str): The question to ask.
            retries (int | None, optional): The number of retries. Defaults to 1.
            max_tokens (int, optional): The maximum number of tokens per message.
                Defaults to 1024.

        Returns:
            The response to the question.
        """
        self.add_message(question)
        return self.run_until_response(retries=retries, max_tokens=max_tokens)

    def add_skill(self, skill: ToolSet) -> None:
        """Add a skill to the conversation.

        Args:
            skill (ToolSet): The skill to add.
        """
        self.skills.add_skill(skill)



"""A basic set of tools."""
from __future__ import annotations

from anthropic.types.beta.tools import ToolParam, ToolUseBlock, ToolUseBlockParam

from anthropic_tools.json_type import JsonType

from ..exceptions import BrokenSchemaError, ToolNotFoundError
from .sets import MutableToolSet
from .tools import AnthropicTool, RawToolResult, ToolResult


class BasicToolSet(MutableToolSet):
    """A basic set of tools."""

    def __init__(self, tools: list[AnthropicTool] | None = None) -> None:
        """Initialize the BasicToolSet.

        Args:
            tools: A list of AnthropicTool objects. Defaults to None.
        """
        self.tools = tools or []

    @property
    def tools_schema(self) -> list[ToolParam]:
        """Get the schema of all tools in the set, in the format the API expects.

        Returns:
            A list of ToolParam objects representing the schema of each tool.
        """
        return [tool.schema for tool in self.tools]

    def run_tool(self, input_data: ToolUseBlock | ToolUseBlockParam) -> ToolResult:
        """Run a tool with the given input data.

        Args:
            input_data: The input data for the tool,
                either as a ToolUseBlock or ToolUseBlockParam object.

        Returns:
            A ToolResult object representing the result of running the tool.

        Raises:
            BrokenSchemaError: If the input data does not match the tool's schema.
        """
        tool = self.find_tool(
            input_data["name"] if isinstance(input_data, dict) else input_data.name
        )
        arguments = (
            input_data["input"] if isinstance(input_data, dict) else input_data.input
        )
        if not isinstance(arguments, dict):
            raise BrokenSchemaError(arguments, tool.schema["input_schema"])
        result = self.get_tool_result(tool, arguments)
        return ToolResult(
            input_data["id"] if isinstance(input_data, dict) else input_data.id,
            tool.name,
            result,
            tool.interpret_as_response,
        )

    def find_tool(self, tool_name: str) -> AnthropicTool:
        """Find a tool by its name.

        Args:
            tool_name: The name of the tool to find.

        Returns:
            The AnthropicTool object with the specified name.

        Raises:
            ToolNotFoundError: If the tool with the specified name is not found.
        """
        for tool in self.tools:
            if tool.name == tool_name:
                return tool
        raise ToolNotFoundError(tool_name)

    def get_tool_result(
        self, tool: AnthropicTool, arguments: dict[str, JsonType]
    ) -> RawToolResult | None:
        """Get the result of running a tool with the given arguments.

        Args:
            tool: The AnthropicTool object representing the tool to run.
            arguments: The arguments for the tool.

        Returns:
            A RawToolResult object representing the result of running the tool,
                or None if the tool does not save the return value.
        """
        result = tool(arguments)
        if tool.save_return:
            return RawToolResult(result, serialize=tool.serialize)
        return None

    def _add_tool(self, tool: AnthropicTool) -> None:
        """Add a tool to the set.

        Args:
            tool: The AnthropicTool object representing the tool to add.
        """
        self.tools.append(tool)

    def _remove_tool(self, name: str) -> None:
        """Remove a tool from the set by its name.

        Args:
            name: The name of the tool to remove.
        """
        self.tools = [f for f in self.tools if f.name != name]
